#include <vector>
#include <stack>
#include <algorithm>
#include <iostream>
#include <bits/stdc++.h>

#pragma once

#define DEBUG(x)
#define EX_DEBUG(y)
#define PARTIAL_EXPERIMENT(z)

class Graph {
    int vertCount;
    int outVertCount;

    std::vector<int> edges, outEdges;
    std::vector<int> orderedEdges, edgeOrder, orderPartSums;

    bool nextMask_walker(int position, int oldPosition, std::vector<bool> &visitMark, std::vector<bool> &retMask0,
                         std::vector<bool> const &extMask, bool sidepath);
    std::stack<std::vector<bool>> maskStack, visitStack;
    std::stack<int> positionStack;

    //Function finds the the smallest metric for the given outFactor.
    int minReducTest( std::vector<bool> const &outFactor);

    //Tries to find a factor of the graph which satisfies the inequality,
    //thus marking this graph as a viable 'reduction'.
    std::vector<bool> reducTest(int boundary, std::vector<bool> const &outFactor);

    public:
    Graph(int vertCount,int outVertCount, std::vector<int> inEdge, std::vector<int> inOutEdges);

    Graph(const Graph &toCopy);

    //-----Main methods------

    //Takes a pair of masks of edges in the even factor, 
    //separate for internal and external edges.
    int evalMetric(std::vector<bool> const &intMask, std::vector<bool> const &extMask);  

    //Returns an outFactor, if it exists, for the smallest possible metric of these two graphs.
    static std::vector<bool> findFactor(Graph &origGraph, Graph &reducedGraph);

    //-----Helper methods------

    //Function either returns a viable mask or an empty vector.
    std::vector<bool> nextMask(std::vector<bool> const &extMask);

    //Expands the existing edges and a mask of deg 1 vertices.
    //Does not modify the outEdge field because that isn't used in the reduction searching algorithm.
    void expandGraph(std::vector<int> const &addEdges, std::vector<int> const &addMask);

    //Uses next_permutation() of <algorithm> to permute the outgoing edges in the graph.
    bool permuteEdges(std::vector<int> const &treeList);

    //Uses next_permutation() of <algorithm> to permute the outgoing edges in the graph.
    //Doesn't employ the optimisation of permuteEdges which treats all reductions as internally orderless.
    //Usable only on graphs which do not have directly connected outgoing vertices.
    bool permuteEdgesExtra();
    
    //-----Miscellaneous methods------

    //Initialises a clean slate for a different mask of outgoing cycles.
    //Only to be used while the stacks are empty.
    void nextMask_reset();

    int getOutVertCount() {return outVertCount;};

    void printGraph() {
        for (int element : edges) {
            std::cout<<element <<" ";
        }
        std::cout<<std::endl;
    };
};

Graph::Graph(int vertCount,int outVertCount,
             std::vector<int> inEdge, std::vector<int> inOutEdges) {
    this->vertCount = vertCount;
    this->outVertCount = outVertCount;

    edges = inEdge;
    outEdges = inOutEdges;
}

Graph::Graph(const Graph &toCopy) {
    this->vertCount = toCopy.vertCount;
    this->outVertCount = toCopy.outVertCount;

    edges = toCopy.edges;
    outEdges = toCopy.outEdges;
}

int Graph::evalMetric(std::vector<bool> const &intMask, std::vector<bool> const &extMask) {
    int triplets = 0, doublets = 0, isolates =0, cycles = 0;
    std::vector<bool> visitMark(vertCount, 0);

    if (intMask.empty() || extMask.empty()) return -404;

    for (int i=0; i<outVertCount; i++) {
        if (extMask[i] && !visitMark[i]) {
            visitMark[i] = true, visitMark[outEdges[i]] = true, cycles++;

            //Jump to the internal vertex connected to i, mark it.
            int neighbor, position = edges[i*3];
            bool moved = true;
                
            while (moved == true) {
                visitMark[position] = true;
                moved = false;
                
                //Find its neighbor (there's always only 1 viable).
                for (int j = 0; j<=2 ; j++) {
                    neighbor = edges[position*3 + j];
                    if ((neighbor != -1) && visitMark[neighbor]) { continue;
                    } else {
                        //Outgoing vertices check for outside routes.
                        if ((position < outVertCount) && (neighbor == -1) &&
                            extMask[position]) {
                            if (visitMark[outEdges[position]]) break;
                            
                            position = outEdges[position];
                            moved = true;
                            break;
                        }
                        if (!intMask[position*3+j]) { continue;
                        } else {
                            position = edges[position*3+j];
                            moved = true;
                            break;
                        }
                    } 
                }
            }
        }
    }

    for (int i=outVertCount; i<vertCount; i++) {
        if (!visitMark[i] && (intMask[i*3] || intMask[i*3+1] || intMask[i*3+2])) {
            visitMark[i] = true, cycles++;

            int neighbor, position = i;
            bool moved = true;
                
            while (moved == true) {
                visitMark[position] = true;
                moved = false;
                
                //Find its neighbor (there's always only 1 viable).
                for (int j = 0; j<=2 ; j++) {
                    neighbor = edges[position*3 + j];
                    if ((neighbor != -1) && visitMark[neighbor]) { 
                        continue;
                    } else {
                        if (!intMask[position*3+j]) { 
                            continue;
                        } else {
                            position = edges[position*3+j];
                            moved = true;
                            break;
                        }
                    } 
                }
            }
        }
    }

    for (int i=outVertCount; i < vertCount; i++) {
        if (!visitMark[i]) isolates++;
        if (edges[i*3+2] == -1) doublets++;
        else                    triplets++;
    }

    return 4*isolates + 8*cycles - 2*doublets - 1*triplets;
}

bool Graph::nextMask_walker(int position, int oldPosition, std::vector<bool> &visitMark, std::vector<bool> &retMask,
                             std::vector<bool> const &extMask, bool sidepath) {
    int neighbor, sanityCheck;
    bool branched, moved = true;

    //Isolated vertex check.
    // if (position == oldPosition) {
    //     DEBUG(std::cout<<"I stepped into an isolated vertex " <<std::endl;)
    //     return false;
    // }

    DEBUG(std::cout<<"I went/returned from " <<position <<" ";)
                
    while (moved == true) {
        branched = false;
        visitMark[position] = true;
        moved = false, sanityCheck = 0;

        //Check if we haven't taken a wrong turn.
        if ((position < outVertCount) && extMask[position]) sanityCheck++;
        for (int i = 0;i<=2 && (edges[position*3 + i] != -1); i++) {
            if (edges[position*3 + i] == oldPosition) retMask[position*3 + i] = true; //The other view of the edge we've just traversed.
            if (visitMark[edges[position*3 + i]] && retMask[position*3 + i]) sanityCheck++; 
        }
        if (sanityCheck >= 3) {
            DEBUG(std::cout<<"and backtracked." <<std::endl;)
            return false; //external loop will try another path
        } if (sanityCheck == 2) {
            DEBUG(std::cout<<"and finished the loop." <<std::endl;)
            return true; //external loop continues.
        }

        for (int j = 0; j<=2 ; j++) {
            neighbor = edges[position*3 + j];
            if ((neighbor == -1) || (!sidepath && visitMark[neighbor]) || (sidepath && retMask[position*3 + j])) { continue;
            } else {
                //New outgoing vertex -> I have to go through it.
                if (neighbor < outVertCount) {
                    if (extMask[neighbor]) {
                        visitMark[neighbor] = true;

                        retMask[position*3 + j] = true; 
                        retMask[neighbor*3] = true;     //Marked from both sides.
                        position = outEdges[neighbor];
                        DEBUG(std::cout<<"through " <<position <<" ";)

                        //Jump inside again (#unrolled optimisation#)
                        visitMark[position] = true;
                        retMask[position*3] = true;
                        oldPosition = position;
                        position = edges[position*3];
                        retMask[position*3] = true; //Again marked from both sides.
                        
                        moved = true;
                        DEBUG(std::cout<<"to " <<position <<" ";)
                        break;
                    } else { continue; }
                //Branch or simply jump forward based on degree.
                } else {
                    if (edges[position*3+2] == -1) {
                        //Degree 2 - single path
                        retMask[position*3+j] = true; 
                        oldPosition = position;
                        position = neighbor;
                        moved = true;
                        DEBUG(std::cout<<"to " <<position <<" ";)
                        break;
                    } else {
                        //Branch - snapshot of positions and the two masks on stack.                        
                        if (!branched && (!sidepath || (edges[position*3] > outVertCount))) {
                            branched = true;
                            std::vector<bool> branchMask = retMask;
                            std::vector<bool> branchVisit = visitMark;

                            branchMask[position*3+j] = true;
                            positionStack.push(position);
                            positionStack.push(edges[position*3+j]);
                            maskStack.push(branchMask);
                            visitStack.push(branchVisit);
                            DEBUG(std::cout<<"mem'd " <<edges[position*3+j] <<" and jumped ";)
                        } else {
                            retMask[position*3+j] = true;
                            oldPosition = position;
                            position = edges[position*3+j];
                            moved = true;
                            DEBUG(std::cout<<"to " <<position <<" ";)
                            break;
                        }                       
                    }
                }
            } 
        }
    }

    DEBUG(std::cout<<"and ended up with no way forward." <<std::endl;)
    return false;
}

std::vector<bool> Graph::nextMask(std::vector<bool> const &extMask) {
    std::vector<bool> visitMark, retMask, emptyMask;
    int position, oldPosition, startPosition = -1;
    
    if (!visitStack.empty()) {
        visitMark = visitStack.top();       visitStack.pop();
        retMask = maskStack.top();          maskStack.pop();
        position = positionStack.top();     positionStack.pop();  
        oldPosition = positionStack.top();  positionStack.pop(); 
    } else { 
        return emptyMask;
    }

    //special case of no prior run
    if (position == 0) startPosition = 0; 

    //Routes starting from the edge, end if unsuccessful.
    for (int i = startPosition; i < outVertCount; i++) {
        //alternative route first.
        if (i == startPosition || (extMask[i] && !visitMark[i])) {
            if (i != -1 && extMask[i] && !visitMark[i]) {
                visitMark[i] = true;
                oldPosition = i;
                position = edges[i*3];
                retMask[i*3] = true;
            }

            //special case as above  *this works, but could be done better. 
            if (position == 0) continue;
        
            DEBUG(std::cout<<"I made a new start from " <<position <<" jumping out of " <<oldPosition <<std::endl;)

            if (nextMask_walker(position, oldPosition, visitMark, retMask, extMask, false)) {
                continue;
            } else {
                if (!visitStack.empty()) {
                    visitMark = visitStack.top();       visitStack.pop();
                    retMask = maskStack.top();          maskStack.pop();
                    position = positionStack.top();     positionStack.pop();  
                    oldPosition = positionStack.top();  positionStack.pop(); 

                    //redo the for cycle.
                    startPosition = -1;
                    i = -2; 
                } else {
                    return emptyMask;
                }
            };            
        }
    }

    DEBUG(std::cout<<"I found what I needed and am now looking for leftover internal cycles" <<std::endl;)

    //Routes looking for straggler cycles, done once everything else finishes.
    std::stack<std::vector<bool>> visitStackBackup, maskStackBackup;
    std::stack<int> positionStackBackup;

    visitStackBackup = visitStack;
    maskStackBackup = maskStack;
    positionStackBackup = positionStack;

    std::vector<bool> stragglers(visitMark), untrodden(retMask);
    std::vector<std::vector<bool>> stragglersFound, pathsTaken;

    std::stack<std::vector<bool>> stragglerStack, pathStack;
    std::stack<int> coordinates;

    visitStack = stragglerStack;
    maskStack = pathStack;
    positionStack = coordinates;
    for (int i = outVertCount; i < vertCount; i++) {
    
        if (!stragglers[i]) {
            stragglers[i] = true;
            oldPosition = -88;
            position = i;
        
            DEBUG(std::cout<<"I made a new start from " <<position <<" jumping out of " <<oldPosition <<std::endl;)

            if (nextMask_walker(position, oldPosition, stragglers, untrodden, extMask, true)) {
                stragglersFound.push_back(stragglers);
                pathsTaken.push_back(untrodden);
            } 
            
            while (!visitStack.empty()) {
                stragglers = visitStack.top();       visitStack.pop();
                untrodden = maskStack.top();          maskStack.pop();
                position = positionStack.top();     positionStack.pop();  
                oldPosition = positionStack.top();  positionStack.pop(); 

                if (nextMask_walker(position, oldPosition, stragglers, untrodden, extMask, true)) {
                    stragglersFound.push_back(stragglers);
                    pathsTaken.push_back(untrodden);
                }
            }

            //Merge the biggest one into retMask and make it the new straggler.
            if (!stragglersFound.empty()) {
                int choice,best = 0;
                for (int j = 0; j < stragglersFound.size(); j++) {
                    std::vector<bool> tmpStragglers = stragglersFound[j];
                    int subtotal = 0; 

                    for (bool ele : tmpStragglers) if (ele) subtotal++;

                    if (subtotal > best) {
                        best = subtotal;
                        choice = j;
                    }
                }

                for (int j = outVertCount*3; j < vertCount*3; j++) {
                    if (pathsTaken[choice][j]) retMask[j] = true;
                }
                stragglers = stragglersFound[choice];

            } else {
                //Or backtrack to the last viable state if not applicable.
                for (int j = outVertCount*3; j < vertCount*3; j++) {
                    untrodden[j] = retMask[j];
                }
            }
        }            
    }

    visitStack = visitStackBackup;
    maskStack = maskStackBackup;
    positionStack = positionStackBackup;


    return retMask;
}

void Graph::nextMask_reset() {
    std::vector<bool> retMask(vertCount * 3, 0);
    std::vector<bool> visitMask(vertCount, 0);

    maskStack.push(retMask);
    visitStack.push(visitMask);
    positionStack.push(-2);     //oldPosition
    positionStack.push(0);      //position
};

std::vector<bool> Graph::reducTest (int boundary, std::vector<bool> const &outFactor) {
    nextMask_reset();
    std::vector<bool> newMask = nextMask(outFactor);
    int currentValue;

    while (!newMask.empty()) {
        currentValue = evalMetric(newMask, outFactor);
        if (currentValue <= boundary) break;
        else newMask = nextMask(outFactor);
    }

    return newMask;
}

int Graph::minReducTest(std::vector<bool> const &outFactor) {
    nextMask_reset();

    std::vector<bool> newMask = nextMask(outFactor);
    int minMetric = INT_MAX, currentValue;

    //pathological case of no viable mask existing.
    if (newMask.empty()) return -3;

    while (!newMask.empty()) {
        currentValue = evalMetric(newMask, outFactor);
        if (currentValue < minMetric) minMetric = currentValue;

        newMask = nextMask(outFactor);
    }

    return minMetric;
};

std::vector<bool> Graph::findFactor(Graph &origGraph, Graph &reducedGraph) {
    std::vector<bool> factorMask, emptyMask;

    //Memorise the old borders in case we need them.
    std::vector<int> oldBorder(origGraph.outEdges);
    std::vector<std::vector<int>> iterStates(1, std::vector<int> (origGraph.outVertCount, -1));

    //Slowly build, test and add new possible pairs for each combination of outside covers. 
    for (int rep = 0; rep < origGraph.outVertCount/2; rep++) {
        std::vector<std::vector<int>> nextIteration = {};

        for (auto element : iterStates) {
            for (int x = 0; x < origGraph.outVertCount; x++) {
                if (element[x] != -1) continue;
                for (int y = x+1; y < origGraph.outVertCount; y++) {
                    if (element[y] != -1) continue;

                    std::vector<int> newElement = element;
                    newElement[x] = y;
                    newElement[y] = x;
                    origGraph.outEdges = newElement;
                    reducedGraph.outEdges = newElement;

                    EX_DEBUG(std::cout<<"===={ I tried out the mask: ";
                        for (auto i = newElement.begin(); i != newElement.end(); ++i) {
                            std::cout << *i << ' ';
                        };
                        std::cout<<" }===="<<std::endl;)

                    std::vector<bool> mask(origGraph.outVertCount, true);
                    for (int i = 0; i < origGraph.outVertCount; i++) 
                        if (newElement[i] == -1) mask[i] = false;

                    int minBound = 0;

                    minBound = reducedGraph.minReducTest(mask);

                    EX_DEBUG(std::cout<<"==[ The score to beat was: " <<minBound <<" ]==" <<std::endl;)

                    nextIteration.push_back(newElement);

                    //No factors of origGraph exist.
                    if (minBound == -3) {
                        PARTIAL_EXPERIMENT(
                        for (auto i = newElement.begin(); i != newElement.end(); ++i) {
                            std::cout << *i << ' ';
                        };
                        std::cout<<std::endl;)
                        continue;
                    } 

                    factorMask = origGraph.reducTest(minBound, mask);

                    if (factorMask.empty()) {
                        PARTIAL_EXPERIMENT(
                        for (auto i = newElement.begin(); i != newElement.end(); ++i) {
                            std::cout << *i << ' ';
                        };
                        std::cout<<std::endl;)

                        EX_DEBUG(std::cout<<"X=X=X=X And we failed X=X=X=X" <<std::endl;)
                        origGraph.outEdges = oldBorder;
                        reducedGraph.outEdges = oldBorder;
                        return emptyMask;
                    }

                    EX_DEBUG(std::cout<<"==[ And we succeeded ";  
                        for (auto i = factorMask.begin(); i != factorMask.end(); ++i) {
                            std::cout << *i << ' ';
                        };
                        std::cout<<" ]==" <<std::endl;)
                }
            }
        }

        iterStates = nextIteration;
    }

    //Restore the old borders.
    origGraph.outEdges = oldBorder;
    reducedGraph.outEdges = oldBorder;

    return factorMask;
};

void Graph::expandGraph(std::vector<int> const &addEdges, std::vector<int> const &addMask) {
    int addOutgoingCount = addMask[1];

    // std::cout<<"Adding ";
    // for (int num : addEdges) {std::cout<<num <<" ";};
    // std::cout<<" The size is " <<addMask[0] <<" and the outgoing are " <<addMask[1] <<std::endl;
    
    //Then merge all the edges, deg 1/original ones first.
    std::vector<int> nEdges ((vertCount*3) + addEdges.size(), -1);
    for (int i = 0; i < outVertCount*3; i+=3) {
        if (edges[i] >= outVertCount)
            nEdges[i] = edges[i] + addOutgoingCount;
        else 
            nEdges[i] = edges[i];
    }
    for (int i = 0; i < addOutgoingCount*3; i+=3) {
        if (addEdges[i] >= addOutgoingCount)
            nEdges[outVertCount*3 + i] = addEdges[i] + vertCount;
        else
            nEdges[outVertCount*3 + i] = addEdges[i] + outVertCount;
    }
    for (int i = outVertCount*3; i < vertCount*3; i++) {
        if (edges[i] >= outVertCount) {
            nEdges[addOutgoingCount*3 + i] = edges[i] + addOutgoingCount;
        }
        else {
            if (edges[i] != -1) nEdges[addOutgoingCount*3 + i] = edges[i];
        } 
    }
    for (int i = addOutgoingCount*3; i < addEdges.size(); i++) {
        if (addEdges[i] >= addOutgoingCount) {
            nEdges[vertCount*3 + i] = addEdges[i] + vertCount;
        }
        else {
            if (addEdges[i] != -1) nEdges[vertCount*3 + i] = addEdges[i] + outVertCount;
        } 
    }

    edges = nEdges;
    outVertCount += addOutgoingCount;
    vertCount += addMask[0];
};

bool Graph::permuteEdges(std::vector<int> const &treeList) {

    //There is nothing to permute at all.
    if (treeList.size() == 1) return false;

    //First init.
    if (orderedEdges.empty()) {
        orderedEdges.resize(outVertCount);
        edgeOrder.resize(outVertCount);
        orderPartSums.resize(treeList.size());
        int count = 0, order = 0;

        for (int i = 0; i < outVertCount; i++) {
            orderedEdges[i] = edges[i*3];
        }
        //populate our selector.
        for (int num : treeList) {
            for (int i = 0; i < num; i++) {
                edgeOrder[count+i] = order;
            }
            orderPartSums[order] = count;
            order++;
            count += num;
        }

        return true;
    } else {
        bool retVal, changed = false;
        std::vector<int> oldEdges(edges);

        do {
            retVal = std::next_permutation(edgeOrder.begin(), edgeOrder.end());
            std::vector<int> connections(vertCount, 0);
            std::vector<int> offsetVals(treeList.size(), 0);
            std::vector<int> offsets(outVertCount, 0);     

            for (int i = 0; i < outVertCount; i++) {
                offsets[i] = offsetVals[edgeOrder[i]];
                offsetVals[edgeOrder[i]]++;
            }

            for (int i = 0; i < outVertCount; i++) {
                int originIndex = orderPartSums[edgeOrder[i]] + offsets[i];
                int pointingAt = orderedEdges[originIndex];

                if (pointingAt < outVertCount) {
                    //We find our shifted pair in the list of pairs.
                    if (pointingAt > originIndex) {
                        int j = i+1;
                        while (edgeOrder[j] != edgeOrder[i]) {j++;}
                        pointingAt = j;
                    } else {
                        int j = i-1;
                        while (edgeOrder[j] != edgeOrder[i]) {j--;}
                        pointingAt = j;
                    }

                    edges[i*3] = pointingAt;
                    edges[pointingAt*3] = i;
                } else {
                    edges[i*3] = pointingAt;
                    edges[pointingAt*3 + connections[pointingAt]] = i;
                    connections[pointingAt]++;
                }
            }

            //Optimise away same shifts.
            for (int i=0; i < outVertCount; i++) {
                if (edges[i*3] != oldEdges[i*3]) changed = true;
            }

        } while (!changed && retVal);

        // printGraph(); 

        return retVal;
    }
};


bool Graph::permuteEdgesExtra() {

    //First init.
    if (orderedEdges.empty()) {
        orderedEdges.resize(outVertCount);
        for (int i = 0; i < outVertCount; i++) {
            orderedEdges[i] = edges[i*3];
        }

        std::sort(orderedEdges.begin(), orderedEdges.end());
        std::vector<int> connections(vertCount, 0);
        for (int i = 0; i < outVertCount; i++) {
            edges[i*3] = orderedEdges[i];
            edges[ orderedEdges[i]*3 + connections[orderedEdges[i]] ] = i;
            connections[orderedEdges[i]]++;
        }

        return true;
    } else {
        bool retVal = std::next_permutation(orderedEdges.begin(), orderedEdges.end());
        std::vector<int> connections(vertCount, 0);

        for (int i = 0; i < outVertCount; i++) {
            edges[i*3] = orderedEdges[i];
            edges[ orderedEdges[i]*3 + connections[orderedEdges[i]] ] = i;
            connections[orderedEdges[i]]++;
        }

        return retVal;
    }
};